
#ifndef __SUBDIV_H__
#define __SUBDIV_H__

#include <vector>

#include "st.h"

/*
	This header contains definitions for structures used to compute
	surface subdivision.
*/

class SubdivFace;
class SubdivVertex;
class SubdivEdge;


class SubdivFace {
	public:
		
		int idx;
		SubdivVertex* verts[3];
		SubdivFace*   adjFaces[3];
		SubdivFace*   childFaces[4];

		SubdivFace() {
			idx = -1;
			adjFaces[0] = adjFaces[1] = adjFaces[2] = NULL;
			childFaces[0] = childFaces[1] = childFaces[2] = childFaces[2] = NULL;
			verts[0] = verts[1] = verts[2] = NULL;
		}


		int VertexIndex(SubdivVertex* vert) {
			for (int i=0; i<3; i++)
				if (verts[i] == vert)
					return i;
			return -1;
		}

		// Returns the face adjacent to the current face that is on
		// the other side of the edge formed by vertex 'vert' and the subsequent
		// vertex in the face
		SubdivFace* NextFace(SubdivVertex* vert) {
			return adjFaces[VertexIndex(vert)];
		}


		// Returns the next vertex in the face after 'vert'
		SubdivVertex* NextVertex(SubdivVertex* vert) {
			return verts[(VertexIndex(vert)+1) % 3]; 
		}


		// Given two vertices from the face, returns the last remaining
		// vertex.
		SubdivVertex* OtherVertex(SubdivVertex* v1, SubdivVertex* v2) {
			for (int i=0; i<3; i++) {
				if (verts[i] != v1 && verts[i] != v2)
				return verts[i];		
			}
			return NULL;
		}
};


class SubdivVertex {
	public:
		int idx;
		STPoint3D    pos; // position of the vertex
		bool         extraordinary; // true if the vertex is extraordinary
		SubdivFace*  adjFace;  // pointer to SOME adjacent face
		SubdivVertex*  childVert; // pointer to child vertex (and EVEN vertex in the next levle of subdivision)
		STVector*    normal;  // point to the vertex normal (only valid for the finest levle of subdivision)

		SubdivVertex() {
			idx = -1;
			pos.x = pos.y = pos.z = 0.0f;
			extraordinary = false;
			adjFace = NULL;
			childVert = NULL;
			normal = NULL;
		}
		
		void GetAdjVertexPositions(std::vector<STVector>& positions);
};



class SubdivEdge {
	public:
		SubdivVertex*  verts[2];   // vertices that define the edfe
		SubdivFace*    adjFace[2]; // faces on either side of the edge
		int adjFace0EdgeNum; // edge number of  this edge in relation to adjFace[0]

		// always store the lower values pointers in verts[0].  This makes
		// comparison and ordering testing either
		SubdivEdge(SubdivVertex* vert0, SubdivVertex* vert1) {

			if (vert0 < vert1) {
				verts[0] = vert0;
				verts[1] = vert1;
			} else {
				verts[0] = vert1;
				verts[1] = vert0;
			}
			
			adjFace[0] = adjFace[1] = NULL;
			adjFace0EdgeNum = -1;
		}

		bool operator<(const SubdivEdge& e2) const {
		if (verts[0] == e2.verts[0])
			return verts[1] < e2.verts[1];
		return verts[0] < e2.verts[0];
	}
};


class SubdivSurface : public STDrawable {

	public:

		SubdivSurface();
		~SubdivSurface();

		void Draw();
		void LoadGeometry(const char* filename);		
		void ComputeSurface(int nLevels);

		void SetColor(STColor& c) {
			mColor = c;
		}

		void DrawWireframe(bool draw) {
			mDrawWireframe = draw;
		}

	private:

		void ProcessBaseGeometry(const std::vector<int>& indices, const std::vector<STPoint3D>& faces);

		std::vector<SubdivFace*> mOrigFaces;
		std::vector<SubdivVertex*> mOrigVerts;

		std::vector<SubdivFace*> mSubdivFaces;
		std::vector<SubdivVertex*> mSubdivVerts;

		std::vector<STVector> mSubdivNormals;

		bool mDrawWireframe;
		STColor mColor;

		// weight: beta
		static float beta(int valence) {
			if (valence == 3)
				return 3.f/16.f;
			else
				return 3.f / (8.f * valence);
		}

		// weight: gamma
		static float gamma(int valence) {
			return 1.f / (valence + 3.f / (8.f * beta(valence)));
		}

		
		// return a new position for the vertex 'vert' given the surrounding vertex positions and beta
		static STPoint3D PositionFromOneRing(SubdivVertex* vert, const std::vector<STVector>& adjPositions, float beta);

};



#endif  // __SUBDIV_H__

