
#include <stdlib.h>

#include "stgl.h"
#include "STFontFace.h"


// Some notes on Freetype:
// Freetype provides many values in 26.6 pixel format = 1/64 of a pixel
// This is a 32 bit fixed point format  


FT_Library* STFontFace::sFTLibrary = NULL;


/**
* Construct instance from font filename and size in points
*/
STFontFace::STFontFace(const char* fontName, int fontSize)
	: mSize(-1), mFTFace(NULL) {

	FT_Error fterr;

	// Need to initialized the freetype2 library
	if (!sFTLibrary) {

		sFTLibrary = new FT_Library();

		fterr = FT_Init_FreeType(sFTLibrary);

		if ( fterr ) {
			fprintf(stderr, "Fatal: Could not initialize freetype library. Error code %d\n", fterr);
			exit(1);
		}
	}

	SetFace(fontName, fontSize);
}


/**
* Destructor just cleans up the Freetype Face object
*/
STFontFace::~STFontFace() {
	if (mFTFace)
		delete mFTFace;
}


/**
* Resets the truetype face and size of the font (in points)
*/
void
STFontFace::SetFace(const char* fontName, int fontSize) {

	FT_Error fterr = FT_New_Face( *sFTLibrary, fontName, 0, &mFTFace);	

	if (fterr) {
		fprintf(stderr, "Warning: Font face %s could not be loaded\n", fontName);
		return;
	}

	// note, the call to setSize() will cause cleanup of existing glyphs
	mSize = -1;
	SetSize(fontSize);
}


/**
* Sets the size of the face in points.
*/
void
STFontFace::SetSize(int fontSize) {

	if (fontSize <= 0) {
		fprintf(stderr, "Warning: Invalid font size: %d.  Defaulting to %dpt\n",
						fontSize, STFontFace::kDefaultFontSize);
		fontSize = STFontFace::kDefaultFontSize;
	}

	// trivial optimization, do nothing if no change
	if (fontSize == mSize)
		return;

	// Express font size in units of 1/64 of a point
	// note: Using default DPI (which happens to be 72 DPI)
	FT_Error fterr = FT_Set_Char_Size( mFTFace, 0, fontSize*64, 0, 0); 

	if (fterr) {
		fprintf(stderr, "Error setting character size of font.\n");
		return;
	}

	mSize = fontSize;

	// We just changed the face size, so all the existing bitmap glyphs are
	// now stale. Destroy all the old bitmaps.  We'll recreate them on demand
	// for the new size

	mCharMap.clear();

	for (size_t i=0; i<mGlyphBitmaps.size(); i++)
		delete [] mGlyphBitmaps[i].data;

	mGlyphBitmaps.clear();

}


/**
* Return size of the face in points
*/
int
STFontFace::GetSize() const {
	return mSize;
}


/**
* Return height of the face in pixels.
*/
float
STFontFace::GetHeight() const {
	
	return (float)mFTFace->size->metrics.height / 64.0f;
}


/**
* Return value of face ascender in pixels
*/
float
STFontFace::GetAscender() const {
	return (float)mFTFace->size->metrics.ascender / 64.0f;
}


/**
* Return value of face descender in pixels
*/
float
STFontFace::GetDescender() const {
	return (float)mFTFace->size->metrics.descender / 64.0f;
}


/**
* Renders a string of text using this face
* Returns the width of the rendered text in pixels (at subpixel accuracy)
*/
float
STFontFace::DrawString(const char* str) {

	if (str == NULL)
		return 0.0f;

	float totalAdvance = 0.0f;

	int strLength = (int)strlen(str);

	// Render glyph for each character in the string
	for (int i=0; i<strLength; i++) {

		char ch = str[i];

		int bitmapIndex = GetBitmapIndex( ch );

		// bitmap for this character hasn't been generated.  Do that now.
		if (bitmapIndex == -1) {

			if (ch == ' ') {
				glBitmap( 0, 0, 0.0f, 0.0f, (float)mSize / 4.0f, 0.0f, (const GLubyte*)0 );
				totalAdvance += (float)mSize / 4.0f;
				continue;
			} else
				bitmapIndex = GenerateBitmap( ch );
		}

		STBitmapGlyph& bitmapGlyph = mGlyphBitmaps[bitmapIndex];

		// Move the raster position by (offsetX, offsetY)
		glBitmap( 0, 0, 0.0f, 0.0f, (float)bitmapGlyph.offsetX, (float)bitmapGlyph.offsetY, (const GLubyte*)0 );

		// Render the glyph to the framebuffer
        glPixelStorei( GL_UNPACK_ROW_LENGTH, bitmapGlyph.pitch * 8);
        glBitmap( bitmapGlyph.width, bitmapGlyph.height, 0.0f, 0.0, 0.0, 0.0, (const GLubyte*)bitmapGlyph.data);

		// Move the raster position to where it should be for the next character.  Note
		// that at present, this code does not take into account font kerning
		glBitmap( 0, 0, 0.0f, 0.0f, (float)-bitmapGlyph.offsetX + bitmapGlyph.advanceX, (float)-bitmapGlyph.offsetY, (const GLubyte*)0 );
		totalAdvance += bitmapGlyph.advanceX;
    }

	return totalAdvance;
}


/**
* Returns the width of the string 'str' in pixels (at subpixel accuracy)
* had the string been rendered in this face.  this is the same logic as in
* DrawString() without issuing the GL commands to actually draw to the screen
*/
float
STFontFace::ComputeWidth(const char* str) {

	if (str == NULL)
		return 0.0f;

	float totalAdvance = 0.0f;

	int strLength = (int)strlen(str);

	for (int i=0; i<strLength; i++) {

		char ch = str[i];

		int bitmapIndex = GetBitmapIndex( ch );

		if (bitmapIndex == -1) {

			if (ch == ' ') {
				glBitmap( 0, 0, 0.0f, 0.0f, (float)mSize / 4.0f, 0.0f, (const GLubyte*)0 );
				totalAdvance += (float)mSize / 4.0f;
				continue;
			} else
				bitmapIndex = GenerateBitmap( ch );

			
		}

		totalAdvance += mGlyphBitmaps[bitmapIndex].advanceX;
    }

	return totalAdvance;
}


/**
* return an index into the mGlyphBitmaps vector that cooresponds to the
* the character 'character'.  Returns -1 if a bitmap for the character
* done not exist
*/
int
STFontFace::GetBitmapIndex(char character) {

	std::map<char, int>::iterator it;

	it = mCharMap.find(character);

	if (it == mCharMap.end())
		return -1;
	else
		return it->second;
}

/**
* Creates a bitmap representation of the glyph for character 'character'
* in this face
*/
int
STFontFace::GenerateBitmap(char character) {
	
	FT_Load_Char( mFTFace, character, FT_LOAD_RENDER | FT_LOAD_MONOCHROME );

	FT_GlyphSlot glyph = mFTFace->glyph;
	FT_Bitmap bitmap = glyph->bitmap;

	if (bitmap.width == 0 || bitmap.rows == 0) {
		// failure: the character might not exist in this face, so attempt to
		// use the glyph for character 0, which is usually rendered as the "missing
		// character" glyph
		//if (FT_Load_Char( mFTFace, 0, FT_LOAD_RENDER | FT_LOAD_MONOCHROME )) {
		//	fprintf(stderr, "Could not load bitmap glyph for char '%c' (value=%d)\n", character, (int)character);
		//}

		return -1;
	}
	

	int bitmapIndex = (int)mGlyphBitmaps.size();
	mCharMap[character] = bitmapIndex;
	mGlyphBitmaps.resize(mGlyphBitmaps.size()+1);

	STBitmapGlyph& bitmapGlyph = mGlyphBitmaps[bitmapIndex];

	unsigned int srcWidth = bitmap.width;
    unsigned int srcHeight = bitmap.rows;
    unsigned int srcPitch = bitmap.pitch;

	// Set all the fields in our structure to represent the glyph

	bitmapGlyph.width = srcWidth;
	bitmapGlyph.height = srcHeight;
	bitmapGlyph.pitch = srcPitch;
	bitmapGlyph.offsetX = glyph->bitmap_left;
	bitmapGlyph.offsetY = glyph->bitmap_top - glyph->bitmap.rows;
	bitmapGlyph.advanceX = (float)glyph->advance.x / 64.0f;

    bitmapGlyph.data = new unsigned char[bitmapGlyph.pitch * bitmapGlyph.height];

	// Like most image formats a rendered bitmap representation of the face's
	// character is going to begin with the top row of the bitmap.  OpenGL 
	// bitmaps begin with the bottom row, so we need to reshuffle the data here

    unsigned char* dest = bitmapGlyph.data + (( bitmapGlyph.height - 1) * bitmapGlyph.pitch);
    unsigned char* src = bitmap.buffer;

    for( unsigned int y = 0; y < srcHeight; ++y)
    {
        memcpy( dest, src, srcPitch);
        dest -= bitmapGlyph.pitch;
        src += srcPitch;
    }

	return bitmapIndex;	
}


