#include "STTransform.h"
#define _USE_MATH_DEFINES
#include <math.h>
#include <string.h>

/**
* Convenience method. Creates and returns a transform corresponding to the
* specified axis-angle rotation
*/
STTransform
STTransform::Rotate(float theta, float x, float y, float z) {
	float c = (float)cosf(float(theta * M_PI / 180.f));
	float s = (float)sinf(float(theta * M_PI / 180.f));

	float len = sqrtf(x*x + y*y + z*z);
	if (len != 0.f) {
		x /= len;
		y /= len;
		z /= len;
	}

	return STTransform( x*x*(1.f-c)+c,   x*y*(1.f-c)-z*s, x*z*(1.f-c)+y*s, 0.f,
						y*x*(1.f-c)+z*s, y*y*(1.f-c)+c,   y*z*(1.f-c)-x*s, 0.f,
						x*z*(1.f-c)-y*s, y*z*(1.f-c)+x*s, z*z*(1.f-c)+c,   0.f,
						0.f,             0.f,             0.f,             1.f);
}

/**
* Convenience method. Creates and returns a transform corresponding to the
* specified translation
*/
STTransform
STTransform::Translate(float xtrans, float ytrans, float ztrans) {
	return STTransform( 1.f, 0.f, 0.f, xtrans,
						0.f, 1.f, 0.f, ytrans,
						0.f, 0.f, 1.f, ztrans,
						0.f, 0.f, 0.f, 1.f );
}

/**
* Convenience method.  Creates and returns a transform corresponding to the
* specified scaling
*/
STTransform 
STTransform::Scale(float xscale, float yscale, float zscale) {
	return STTransform( xscale, 0.f,    0.f,    0.f,
						0.f,    yscale, 0.f,    0.f,
						0.f,    0.f,    zscale, 0.f,
						0.f,    0.f,    0.f,    1.f );
}

/**
* Initialize transform to the identity
*/
STTransform::STTransform() {
	memset(this->m, 0, 16*sizeof(float));
	this->m[0][0] = 1.f;
	this->m[1][1] = 1.f;
	this->m[2][2] = 1.f;
	this->m[3][3] = 1.f;
}

/**
* Initialize transform using a matrix given as individual floats
*/
STTransform::STTransform(const float _m00, const float _m10, const float _m20, const float _m30,
							const float _m01, const float _m11, const float _m21, const float _m31,
							const float _m02, const float _m12, const float _m22, const float _m32,
							const float _m03, const float _m13, const float _m23, const float _m33) {
	this->m[0][0] = _m00;
	this->m[0][1] = _m01;
	this->m[0][2] = _m02;
	this->m[0][3] = _m03;

	this->m[1][0] = _m10;
	this->m[1][1] = _m11;
	this->m[1][2] = _m12;
	this->m[1][3] = _m13;

	this->m[2][0] = _m20;
	this->m[2][1] = _m21;
	this->m[2][2] = _m22;
	this->m[2][3] = _m23;

	this->m[3][0] = _m30;
	this->m[3][1] = _m31;
	this->m[3][2] = _m32;
	this->m[3][3] = _m33;
}

/**
* Initialize transform from another transform
*/
STTransform::STTransform(const STTransform& xform) {
	memcpy(this->m, xform.m, 16*sizeof(float));
}

/**
* Initialize transform using a matrix given as an array
*/
STTransform::STTransform(const float _m[4][4]) {
	memcpy(this->m, _m, 16*sizeof(float));
}

/* The matrix is stored in GL column major format,
 * so its safe to simply pass back a pointer to 
 * the first element.
 */
float* STTransform::GetGLMatrix() {
	return &m[0][0];
}

/* Matrix multiplication, implemented in the obvious way. */
STTransform STTransform::operator*(const STTransform& rhs) {
	float resultm[4][4];
	int i, j, k;

	for(i = 0; i < 4; i++) {
		for(j = 0; j < 4; j++) {
			resultm[j][i] = 0.f;
			for(k = 0; k < 4; k++) {
				resultm[j][i] += this->m[k][i] * rhs.m[j][k];
			}
		}
	}

	return STTransform(resultm);
}
